# Parallel Firewall

## Objectives

- Learn how to design and implement parallel programs
- Get experienced at utilizing the POSIX threading API
- Learn how to convert a serial program into a parallel one

## Statement

A firewall is a program that checks network packets against a series of filters which provide a decision regarding dropping or allowing the packets to continue to their intended destination.

**In a real setup**, the network card will receive real packets (e.g. packets having [`Ethernet`](https://en.wikipedia.org/wiki/Ethernet_frame), [`IP`](https://en.wikipedia.org/wiki/IPv4) headers plus payload) from the network and will send them to the firewall for processing.
The firewall will decide if the packets are to be dropped or not and, if not, passes them further.

**In this project**, instead of real network packets, we'll deal with made up packets consisting of a made up source (a number), a made up destination (also a number), a timestamp (also a number) and some payload.
And instead of the network card providing the packets, we'll have a **producer thread** creating these packets.

The created packets will be inserted into a [`circular buffer`](https://en.wikipedia.org/wiki/Circular_buffer), out of which **consumer threads** (which implement the firewall logic) will take packets and process them in order to decide whether they advance to the destination.

The result of this processing is a log file in which the firewall will record the decision taken (PASS or DROP) for each packet, along with other information such as timestamp.

In this project I will:

- implement the circular buffer, along with synchronization mechanisms for it to work in a multithreaded program

- implement the consumer threads, which consume packets and process them

- provide the log file containing the result of the packet processing

## Code

The code consists of the directories:

- `src/` contains the code for the parallelized firewall and the serial code in `src/serial.c`.

- `utils/` contains utility files used for debugging and logging.

- `tests/` contains tests used to validate the project.

### Firewall Threads

In order to parallelize the firewall we have to distribute the packets to multiple threads.
The packets will be added to a shared data structure (visible to all threads) by a `producer` thread and processed by multiple `consumer` threads.
Each `consumer` thread picks a packet from the shared data structure, checks it against the filter function and writes the packet hash together with the drop/accept decision to a log file.
`consumer` threads stop waiting for new packets from the `producer` thread and exit when the `producer` thread closes the connection to the shared data structure.

The `consumer` threads **does not do any form of busy waiting**.
When there are new packets that need to be handled, the `consumer` threads will be **notified**.

**The number of consumer threads will be passed as the 3rd command-line argument**

### Ring Buffers

A ring buffer (or a circular buffer) is a data structure that stores its elements in a circular fixed size array.
One of the advantages of using such a data structure as opposed to an array is that it acts as a FIFO, without the overhead of moving the elements to the left as they are consumed.
Thus, the shared ring buffer offers the following fields:

- `write_pos` index in the buffer used by the `producer` thread for appending new packets.
- `read_pos`  index in the buffer used by the `consumer` threads to pick packets.
- `cap` the size of the internal buffer.
- `data` pointer to the internal buffer.

Apart from these fields I added synchronization primitives in order to allow multiple threads to access the ring buffer in a deterministic manner.

- `ring_buffer_init()`: initialize the ring buffer (allocate memory and synchronization primitives).
- `ring_buffer_enqueue()`: add elements to the ring buffer.
- `ring_buffer_dequeue()`: remove elements from the ring buffer.
- `ring_buffer_destroy()`: free up the memory used by the ring_buffer.
- `ring_buffer_stop()`: finish up using the ring buffer for the calling thread.

### Log File

The output of the firewall will be a log file with the rows containing the firewall's decision, the hash of the packet and its timestamp.
The actual format can be found in the serial implementation (at `src/serial.c`).

When processing the packets in parallel the threads will finish up the work in a non deterministic order.
The packet processing functions are already implemented in `src/packet.c`

We would like the logs to be sorted by the packet timestamp, the order that they came in from the producer.
Thus, the `consumers` should insert the packet information to the log file such as the result is ordered by timestamp.
The printing format can be found in `./src/serial.c`

### Building

To build both the serial and the parallel versions, run `make` in the `src/` directory:

```console
student@so:~/.../content/assignments/parallel-firewall$ cd src/

student@so:~/.../assignments/parallel-firewall/src$ make
```

That will create the `serial` and `firewall` binaries.

### Testing

Input tests cases are located in `tests/in/` and are generated by the checker.
The expected results are generated by the checker while running the serial implementation.
If you want to run a single test, use the below commands while in the `src/` directory:

```console
student@so:~/.../assignments/parallel-firewall/src$ ./firewall ../tests/in/test_<num_packets>.in <output_file> <number_of_consumers>
```