// SPDX-License-Identifier: BSD-3-Clause

#include <pthread.h>
#include <fcntl.h>
#include <unistd.h>

#include "consumer.h"
#include "ring_buffer.h"
#include "packet.h"
#include "utils.h"

extern timestamp_list list;

void consumer_thread(so_consumer_ctx_t *ctx)
{
	while (1) {
		// iau un pachet din ring buffer
		char pack_buffer[PKT_SZ];

		ssize_t deque_status = ring_buffer_dequeue(ctx->producer_rb, pack_buffer, PKT_SZ);

		if (deque_status < 0)
			break;

		struct so_packet_t *pkt = (struct so_packet_t *)pack_buffer;

		// procesez pachetul
		int action = process_packet(pkt);
		unsigned long hash = packet_hash(pkt);
		unsigned long timestamp = pkt->hdr.timestamp;

		char out_buf[PKT_SZ];

		int len = snprintf(out_buf, 256, "%s %016lx %lu\n",
			RES_TO_STR(action), hash, timestamp);

		// scriu in fisier
		pthread_mutex_lock(&list.lock);

		if (list.head == NULL && list.stop == 1) {
			pthread_mutex_unlock(&list.lock);
			break;
		}

		// astept sa ajung la timestamp-ul pachetului ca sa le am in ordine in fisier
		while (list.head != NULL && list.head->timestamp != timestamp)
			pthread_cond_wait(&list.wait_packet, &list.lock);

		// daca thread-ul curent nu mai asteapta, le zic celorlalte thread-uri sa verifice
		// daca timestamp-ul lor este cel care trebuie scris
		pthread_cond_broadcast(&list.wait_packet);

		// scriu in fisier si scot din lista de pachete sortate, timestamp-ul scris
		// cum lista e sortata, cel scris e chiar capul listei
		if (list.head != NULL && timestamp == list.head->timestamp) {
			pthread_mutex_lock(&ctx->file_lock);
			write(ctx->out_fd, out_buf, len);
			pthread_mutex_unlock(&ctx->file_lock);

			remove_from_list(&list);
		}

		pthread_mutex_unlock(&list.lock);
	}
}

int create_consumers(pthread_t *tids,
					 int num_consumers,
					 struct so_ring_buffer_t *rb,
					 const char *out_filename)
{
	int out_fd = open(out_filename, O_WRONLY|O_CREAT|O_TRUNC, 0666);
	so_consumer_ctx_t *ctx = malloc(sizeof(so_consumer_ctx_t));

	ctx->producer_rb = rb;
	ctx->out_fd = out_fd;
	pthread_mutex_init(&ctx->file_lock, NULL);

	for (int i = 0; i < num_consumers; i++)
		pthread_create(&tids[i], NULL, (void *)consumer_thread, ctx);

	return num_consumers;
}
