// SPDX-License-Identifier: BSD-3-Clause

#include "ring_buffer.h"
#include "packet.h"
#include "utils.h"

timestamp_list list;

int ring_buffer_init(so_ring_buffer_t *ring, size_t cap)
{
	char *data = malloc(cap);

	DIE(data == NULL, "malloc failed");

	ring->data = data;
	ring->read_pos = 0;
	ring->write_pos = 0;
	ring->len = 0;
	ring->cap = cap;
	ring->stop = 0;

	pthread_mutex_init(&ring->lock, NULL);
	pthread_cond_init(&ring->full_cond, NULL);
	pthread_cond_init(&ring->empty_cond, NULL);

	// initializare lista de timestamp-uri
	init_list();

	return 0;
}

ssize_t ring_buffer_enqueue(so_ring_buffer_t *ring, void *data, size_t size)
{
	pthread_mutex_lock(&ring->lock);

	// astept sa se elibereze spatiu in ring buffer
	while ((ring->write_pos + size) % ring->cap == ring->read_pos && !ring->stop)
		pthread_cond_wait(&ring->full_cond, &ring->lock);

	memcpy(ring->data + ring->write_pos, data, size);
	ring->write_pos = (ring->write_pos + size) % ring->cap;
	ring->len += size;

	// semnalez ca am adaugat date in ring buffer si ca sigur nu mai este gol
	pthread_cond_signal(&ring->empty_cond);

	// am reusit sa adaug un pachet in ring buffer si atunci adaug timestamp-ul in lista
	struct so_packet_t *pkt = (struct so_packet_t *)data;
	unsigned long timestamp = pkt->hdr.timestamp;

	// adaug timestamp-ul in lista simplu inlantuita sortata
	add_to_list(timestamp);

	pthread_mutex_unlock(&ring->lock);

	return size;
}

ssize_t ring_buffer_dequeue(so_ring_buffer_t *ring, void *data, size_t size)
{
	pthread_mutex_lock(&ring->lock);

	// astept sa se adauge date in ring buffer
	while (ring->read_pos == ring->write_pos && !ring->stop)
		pthread_cond_wait(&ring->empty_cond, &ring->lock);

	if (ring->stop == 1 && ring->len == 0) {
		pthread_mutex_unlock(&ring->lock);
		return -1;
	}

	memcpy(data, ring->data + ring->read_pos, size);
	ring->read_pos = (ring->read_pos + size) % ring->cap;
	ring->len -= size;

	// semnalez ca am scos date din ring buffer si nu mai este cu siguranta plin
	pthread_cond_signal(&ring->full_cond);
	pthread_mutex_unlock(&ring->lock);

	return size;
}

void ring_buffer_destroy(so_ring_buffer_t *ring)
{
	free(ring->data);
	pthread_mutex_destroy(&ring->lock);
	pthread_cond_destroy(&ring->empty_cond);
	pthread_cond_destroy(&ring->full_cond);

	pthread_mutex_destroy(&list.lock);
	pthread_cond_destroy(&list.wait_packet);
}

void ring_buffer_stop(so_ring_buffer_t *ring)
{
	pthread_mutex_lock(&ring->lock);
	ring->stop = 1;

	pthread_cond_broadcast(&ring->empty_cond);
	pthread_cond_broadcast(&ring->full_cond);

	pthread_mutex_unlock(&ring->lock);

	// pentru lista de timestamp-uri
	pthread_mutex_lock(&list.lock);
	list.stop = 1;

	pthread_cond_broadcast(&list.wait_packet);
	pthread_mutex_unlock(&list.lock);
}

void init_list(void)
{
	list.head = NULL;
	list.tail = NULL;
	pthread_mutex_init(&list.lock, NULL);
	pthread_cond_init(&list.wait_packet, NULL);
}

void add_to_list(unsigned long timestamp)
{
	pthread_mutex_lock(&list.lock);

	list_node *new_node = malloc(sizeof(list_node));

	DIE(new_node == NULL, "malloc failed");

	new_node->timestamp = timestamp;
	new_node->next = NULL;

	// cum pachetele sunt create de producer in ordine crescatoare
	// a timestamp-urilor, adaug in lista la final, acestea fiind deja sortate
	if (list.head == NULL) {
		list.head = new_node;
		list.tail = new_node;
	} else {
		list_node *current = list.tail;

		current->next = new_node;
		list.tail = new_node;
	}

	pthread_mutex_unlock(&list.lock);
}

void remove_from_list(void)
{
	// scot primul element din lista
	list_node *current = list.head;

	if (current != NULL)
		list.head = current->next;

	free(current);
}
